/**********************************************************************
 * ObjectTeams/Java Runtime Environment.
 * This file is part of "Object Teams Development Tooling"-Software
 * 
 * Copyright 2006-2008 Berlin Institute of Technology, Germany.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * $Id: BaseMethodTransformation.java 18986 2008-09-18 14:55:51Z stephan $
 * 
 * Please visit http://www.objectteams.org for updates and contact.
 * 
 * Contributors:
 * Berlin Institute of Technology - Initial API and implementation
 * Authors: 
 * - Christine Hundt (2006)
 * - Stephan Herrmann (2008)
 **********************************************************************/
package org.objectteams;

import java.util.HashSet;
import java.util.WeakHashMap;


/**
 * This class is for internal use, only.
 * 
 * Maintain information about existing threads as to manage
 * team activation per thread vs. globally.
 * 
 * @author resix
 */
public class TeamThreadManager {

	private static Object token = new Object();
	
	private static HashSet<Team> globalActiveTeams = new HashSet<Team>();
	private static WeakHashMap<Team,Object> teamsWithActivationInheritance = new WeakHashMap<Team,Object>();
	private static HashSet<Thread> existingThreads = new HashSet<Thread>();
	
	public static boolean newThreadStarted(boolean isMain, Thread parent) {
		if (!isMain && (new Exception().getStackTrace().length > 3))
			return false;
		// workaround for application hang on Mac OS with Apple JVM:
		Thread currentThread = Thread.currentThread();
		if (System.getProperty("os.name").startsWith("Mac"))
			if (currentThread.getName().equals("AWT-Shutdown"))
				return false;

		Team[] globalTeams;
		Team[] inheritableTeams;
		synchronized (TeamThreadManager.class) {
			existingThreads.add(currentThread);
			
			globalTeams = globalActiveTeams.toArray(new Team[globalActiveTeams.size()]);
			inheritableTeams = teamsWithActivationInheritance.keySet().toArray(new Team[teamsWithActivationInheritance.size()]);
		}
		// activate teams outside synchronized block:
		for (Team t : globalTeams)			
			t.activate(currentThread); // small version? global -> already registered...!
		if (parent != null)
			for (Team t : inheritableTeams)
				if (t.internalIsActiveSpecificallyFor(parent))
					t.activate(currentThread); // pass activation from parent to child thread
		return true;
	}
	public static void threadEnded() {
		Team[] teamsToDeactivate = internalThreadEnded();
		// + remove per thread activation:
		for (Team t : teamsToDeactivate) 
			//t.deactivate(Thread.currentThread()); // small version?
			t.deactivateForEndedThread(Thread.currentThread());			
	}
	private synchronized static Team[] internalThreadEnded() {
		existingThreads.remove(Thread.currentThread());
		// fetch all global active teams for deactivation:
		return globalActiveTeams.toArray(new Team[globalActiveTeams.size()]);
	}
	
	public synchronized static void addGlobalActiveTeam(Team t) {
		globalActiveTeams.add(t);
	}
	
	public synchronized static void removeGlobalActiveTeam(Team t) {
		globalActiveTeams.remove(t);
	}
	
	public static HashSet<Thread> getExistingThreads() {
		return existingThreads;
	}
	public static void registerTeamForActivationInheritance(Team aTeam) {
		teamsWithActivationInheritance.put(aTeam,token);
	}
	public static void unRegisterTeamForActivationInheritance(Team aTeam) {
		teamsWithActivationInheritance.remove(aTeam);
	}

}
