/**********************************************************************
 * ObjectTeams/Java Runtime Environment.
 * This file is part of "Object Teams Development Tooling"-Software
 * 
 * Copyright 2005-2008 Berlin Institute of Technology, Germany.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * $Id: JManglerEnhancer.java 20107 2009-05-11 11:49:24Z stephan $
 * 
 * Please visit http://www.objectteams.org for updates and contact.
 * 
 * Contributors:
 * Berlin Institute of Technology - Initial API and implementation
 * Authors/Contributors: 
 * - Christine Hundt (2005-2006)
 * - Jürgen Widiker  (2005)
 * - Stephan Herrmann (2008)
 **********************************************************************/
package org.objectteams.transformer.jmangler;

import org.cs3.jmangler.bootstrap.ClassPathClassLoader;
import org.cs3.jmangler.supplier.ClassSetException;
import org.cs3.jmangler.supplier.UnextendableClassSet;
import org.cs3.jmangler.tau.ExtensionSet;
import org.cs3.jmangler.tau.InterfaceTransformerComponent;
import org.objectteams.transformer.ClassEnhancer;
import org.objectteams.transformer.ObjectTeamsTransformation;

import de.fub.bytecode.Constants;
import de.fub.bytecode.classfile.Field;
import de.fub.bytecode.classfile.Method;
import de.fub.bytecode.generic.ClassGen;
import de.fub.bytecode.generic.ConstantPoolGen;


/**
* This class implements the ClassEnhancer interface with the JMangler specific behavior.
*  
* @author  Christine Hundt, Juergen Widiker
*/

public class JManglerEnhancer implements ClassEnhancer, Constants {
	
	private ExtensionSet es;
	private InterfaceTransformerComponent initiator;
	private UnextendableClassSet classSet;
	
	/**
	 * Constructor
	 * @param initiator
	 * @param es
	 * @param classSet
	 */
	public JManglerEnhancer(InterfaceTransformerComponent initiator,
							ExtensionSet es,
							UnextendableClassSet classSet) {
		this.initiator = initiator;
		this.es = es;
		this.classSet = classSet;
	}

	/* (non-Javadoc)
	 * @see org.objectteams.transformer.ClassEnhancer#addImplements(java.lang.String, de.fub.bytecode.generic.ClassGen)
	 */
	public void addImplements(String interfaceName, ClassGen cg) {
		es.addInterfaceToClass(initiator, cg.getClassName(), interfaceName);		
	}

	/* (non-Javadoc)
	 * @see org.objectteams.transformer.common.ClassEnhancer#addMethod(de.fub.bytecode.classfile.Method, de.fub.bytecode.generic.ClassGen)
	 */
	public void addMethod(Method m, ClassGen cg) {
		es.addMethod(initiator,cg.getClassName(), m);
	}

	public void addOrReplaceMethod(Method method, ClassGen cg) {
		Method existingMethod = cg.containsMethod(method.getName(), method.getSignature());
		if (existingMethod == null)
			addMethod(method, cg);
		else
			cg.replaceMethod(existingMethod, method);
	}
	
	/* (non-Javadoc)
	 * @see org.objectteams.transformer.common.ClassEnhancer#addField(de.fub.bytecode.classfile.Field, de.fub.bytecode.generic.ClassGen)
	 */
	public void addField(Field f, ClassGen cg) {
		es.addField(initiator,cg.getClassName(),f);
	}

	/* (non-Javadoc)
	 * @see org.objectteams.transformer.common.ClassEnhancer#loadClass(java.lang.String)
	 */
	public void loadClass(String className, ObjectTeamsTransformation client) {
		es.loadClass(initiator, className);
		// done, no need to call back to the client
	}

	/* (non-Javadoc)
	 * @see org.objectteams.transformer.common.ClassEnhancer#decapsulateMethod(de.fub.bytecode.classfile.Method, java.lang.String, de.fub.bytecode.generic.ConstantPoolGen)
	 */
	public void decapsulateMethod(Method m, ClassGen cg, String packageName, ConstantPoolGen cpg) {
		String className = cg.getClassName();
		int flags = m.getAccessFlags();
		String methodName = m.getName();
		if ((flags & ACC_PRIVATE) != 0)
			es.removeMethodModifier(initiator,
									className, 
									methodName,
									m.getSignature(),
									ACC_PRIVATE);
		if ((flags & ACC_PROTECTED) != 0)
			es.removeMethodModifier(initiator,
									className, 
									methodName,
									m.getSignature(),
									ACC_PROTECTED);
		if ((flags & ACC_PUBLIC) == 0) {			 
			es.addMethodModifier   (initiator,
									className, 
									methodName,
									m.getSignature(),
									ACC_PUBLIC);
//			printLogMessage("Adjusting from "
//							+Utility.accessToString(flags)
//							+" to public:\n\t"
//							+className+"."+m);
			if (!packageName.equals("NO_PACKAGE")) 
				checkSeal(packageName, classSet, className);
		}
	}
	
	private static void checkSeal(String package_name,
			UnextendableClassSet cs,
			String class_name) 
	{
		ClassLoader cl = null;
		try {
			cl = cs.getResponsibleClassLoader(class_name);
		} catch (ClassSetException e) {
			throw new IllegalAccessError("OT/J decapsulation: " + e);
		}
		// need to access as ClassPathClassLoader to call getPackage,
		// which is protected in ClassLoader.
		ClassPathClassLoader cpcl = (ClassPathClassLoader)cl;
		Package pckg = null;
		try {
			pckg = cpcl.getPackage(package_name);
		} catch (IllegalAccessError e) {
			System.err.println("*** OTRE Warning: cannot check sealing of package "
								+ package_name + ".");
			System.err.println("(JMangler patch for getPackage() not installed.)");
			System.err.println("original error: " + e);
			return;
		}
		if ( (pckg != null) && pckg.isSealed()) 
			throw new IllegalAccessError(
					"OT/J callout binding:\n"
					+ "Trying to break sealed " + pckg);
	}

}
